import warnings

from ..base import offsets, utils  # noqa


warnings.warn('the vlbi_base module has been deprecated in baseband 4.0. '
              'Use the base module instead.', DeprecationWarning)
